package ca.recrown.islandsurvivalcraft.world.generation;

import java.util.Arrays;

import org.bukkit.World;
import org.bukkit.block.Biome;

import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateTargetValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.DepthFirstSearch;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;

public class BiomePerIslandGenerator implements IslandBiomeGenerator, CoordinateTargetValidatable {
    private boolean initialized;
    private IslandWorldMapper islandLocator;
    private BiomeSelector biomeSelector;
    private TemperatureMapGenerator temperatureMapGenerator;
    private World world;
    private DepthFirstSearch dfs;
    private Biome mainBiome;
    private Biome shoreBiome;
    private float temperature;

    private int chunkX, chunkZ;
    private Biome[][] chunkCache;

    public BiomePerIslandGenerator() {
        this.temperatureMapGenerator = new TemperatureMapGenerator();
        chunkCache = new Biome[16][16];
    }
    
    @Override
    public void initialize(World world, IslandWorldMapper mapGenerator, BiomeSelector biomeSelector) {
        if (initialized) throw new IllegalStateException("Biome generator already initialized.");
        initialized = true;
        this.islandLocator = mapGenerator;
        this.biomeSelector = biomeSelector;
        this.world = world;
        this.temperatureMapGenerator.setSeed(world.getSeed());
        dfs = new DepthFirstSearch(32, islandLocator);
    }

    @Override
    public boolean isCoordinateTarget(int x, int y) {
        //Should be ran for a island until finding biome info or until it looked for it's entirety and couldn't find it. This is one island.

        //First check if the chunk we're looking at is generated.
        if (world.isChunkGenerated(x/16, y/16)) {
            //If so, we know that the biomes have already been assigned, and can therefore use it.
            Biome foundBiome = world.getBiome(x, 0, y);
            if (islandLocator.isShore(x, y)) {
                this.shoreBiome = foundBiome;
            } else {
                this.mainBiome = foundBiome;
            }
            if (mainBiome != null && shoreBiome != null) {
                return true;
            }
        } else
        
        //If the same chunk, then do the same thing, except use the chunk biome cache instead to look for it.
        if (x/16 == chunkX && y/16 == chunkZ) {
            Biome foundBiome = chunkCache[Math.abs(x % 16)][Math.abs(y % 16)];
            if (foundBiome == null) return false;
            if (islandLocator.isShore(x, y)) {
                this.shoreBiome = foundBiome;
            } else {
                this.mainBiome = foundBiome;
            }
            if (mainBiome != null && shoreBiome != null) {
                return true;
            }
        }
        return false;
    }


    @Override
    public Biome GenerateBiome(int chunkX, int chunkZ, int localX, int localZ) {
        int worldX = chunkX * 16 + localX;
        int worldZ = chunkZ * 16 + localZ;

        //Clear chunk cache if different chunk
        if (chunkX != this.chunkX || chunkZ != this.chunkZ) {
            this.chunkX = chunkX;
            this.chunkZ = chunkZ;
            for (int i = 0; i < chunkCache.length; i++) {
                Arrays.fill(chunkCache[i], null);
            }
        }

        if (islandLocator.isIsland(worldX, worldZ)) {
            if (mainBiome == null && shoreBiome == null) {
                dfs.setStartPosition(worldX, worldZ);
                if (!dfs.findTarget(this)) {
                    temperature = temperatureMapGenerator.getTemperature(worldX, worldZ);
                    mainBiome = biomeSelector.getLandBiome(temperature);
                    shoreBiome = biomeSelector.getShoreBiome(mainBiome, temperature);
                }
            }
        } else {
            mainBiome = null;
            shoreBiome = null;
        }

        
        Biome designatedBiome = mainBiome;
        if (mainBiome == null) {
            designatedBiome = biomeSelector.getOceanBiome(temperatureMapGenerator.getTemperature(worldX, worldZ));
        } else if (islandLocator.isShore(worldX, worldZ)) {
                designatedBiome = shoreBiome;
        }

        chunkCache[localX][localZ] = designatedBiome;
        return designatedBiome;
    }

    @Override
    public IslandBiomeGenerator getInstance() {
        return new BiomePerIslandGenerator();
    }
}